/*
 * glgraph.c - A port of some functions of xgraph to opengl+glut
 *
 * Tsutomu Sakurai <tsakurai@rimath.saitama-u.ac.jp>
 * 1999.03
 */

/*
 * xgraph copyrigh notice:
 *
 > --- From copyright.h ------------------------------------------------
 >
 > xgraph - program to graphically display numerical data
 >
 > David Harrison
 > University of California, Berkeley
 > 1989
 >
 > Copyright (c) 1988, 1989, Regents of the University of California.
 > All rights reserved.
 >
 > Use and copying of this software and preparation of derivative works
 > based upon this software are permitted.  However, any distribution of
 > this software or derivative works must include the above copyright
 > notice.
 >
 > This software is made available AS IS, and neither the Electronics
 > Research Laboratory or the University of California make any
 > warranty about the software, its performance or its conformity to
 > any specification.
 > --------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <float.h>
#include <math.h>
#include <GL/glut.h>
#include <string>
#include "glgraph.h"
#include "copyright.h"

/* void SaveImageAs(const char *fname) */
#include "bmp.c" 

/*****************************************
 * glgraph global staff 
 ----------------------------------------*/
/* default window size */
#define WIDTH  560
#define HEIGHT 560

/* Main object of glgraph */
DataSet PlotData[MAXSETS];   // MAXSETS 64

/* Global Parameters */
double llx, urx, lly, ury;   /* Bounding box of all data */

/*******************************
 * glut bitmap fonts 
 ------------------------------*/ 
struct glut_bitmap_font {
    char *name;
    void *font;
} fontmap[] = {
{"8x13",           GLUT_BITMAP_8_BY_13},
{"9x15",           GLUT_BITMAP_9_BY_15},
{"helvetica-10",   GLUT_BITMAP_HELVETICA_10},
{"helvetica-12",   GLUT_BITMAP_HELVETICA_12},
{"helvetica-18",   GLUT_BITMAP_HELVETICA_18},
{"times-roman-10", GLUT_BITMAP_TIMES_ROMAN_10},
{"times-roman-24", GLUT_BITMAP_TIMES_ROMAN_24},
{NULL, NULL},
};

void *mapfont(char *fontname)
{
    int i;
    for (i = 0; fontmap[i].name; i++) 
	if (strcmp(fontname, fontmap[i].name) == 0) break;
    if (fontmap[i].name)
	return fontmap[i].font;
    else {
	fprintf(stderr, "\nillegal font name: %s\n\n", fontname); exit (1);
    }
}

void *setFont(char *fontname) {
    static void	*curfont = GLUT_BITMAP_HELVETICA_12;
    return (fontname ? curfont = mapfont(fontname) : curfont);
}

void *getFont(void)      { return setFont((char *)0); };

/*****************************************
 * for parameter setting (params.c)
 ----------------------------------------*/
extern void pm_set(const char *Key, char  *val);
extern void pm_set(const char *Key, int    val);
extern void pm_set(const char *Key, double val);

/*****************************************
 * Window Info
 ----------------------------------------*/
#define MAX(a,b)     ((a > b) ? a : b)
#define MIN(a,b)     ((a < b) ? a : b)
    
int strwidth(char * str, char *fontname = 0)
{
    const int default_char_width = 6;
    void *font = (fontname) ? setFont(fontname): getFont();
    int swidth = 0;
    if (glutGetWindow() != 0)
	while (*str) { swidth += glutBitmapWidth(font, *str++); }
    else
	swidth = strlen(str)*default_char_width;
    return swidth;
}

int strheight(char * str, char *fontname = 0)
{
    void *font = (fontname) ? setFont(fontname): getFont();
    return
	(font == GLUT_BITMAP_HELVETICA_10)   ?  7:
	(font == GLUT_BITMAP_HELVETICA_12)   ?  8:
	(font == GLUT_BITMAP_HELVETICA_18)   ? 12:
	(font == GLUT_BITMAP_TIMES_ROMAN_10) ?  6:
	(font == GLUT_BITMAP_TIMES_ROMAN_24) ? 16:
	(font == GLUT_BITMAP_8_BY_13)        ?  8:
	(font == GLUT_BITMAP_9_BY_15)        ? 10:
	/* else */                              8;
}

LocalWin *getWinInfo(void)
{
    extern double llx, urx, lly, ury;
    static LocalWin wi;
    char  label[128], *XUnitText = pmString("XUnitText");
    char *LabelFont = pmString("LabelFont");
    int	  expX;
    const int
	xpad = 10,
	ypad = 10;

    /* read area size */
    if (glutGetWindow() != 0) { 
	wi.area_w = glutGet((GLenum) GLUT_WINDOW_WIDTH); 
	wi.area_h = glutGet((GLenum) GLUT_WINDOW_HEIGHT);
    } else {
	wi.area_w = pmInt("Width");
	wi.area_h = pmInt("Height");
    }
    
    /* Setup Parameters */
    wi.UsrOrgX = llx -.05*(urx-llx);
    wi.UsrOppX = urx +.05*(urx-llx);
    wi.UsrOrgY = lly -.05*(ury-lly);
    wi.UsrOppY = ury +.05*(ury-lly);
	
    if ((expX = (int)floor(log10(MAX(fabs(wi.UsrOrgX),fabs(wi.UsrOppX)))/3)*3))
	sprintf(label, "%s x 10%d", XUnitText, expX);
    else 
	sprintf(label, "%s", XUnitText);
    int lwidth = MAX(20, strwidth(label, LabelFont));
    for (int idx = 0; idx < MAXSETS; idx++) 
	if (PlotData[idx].list) {
	    lwidth = MAX(lwidth, strwidth(PlotData[idx].setName, LabelFont));
	}
    pm_set("LegendWidth", lwidth);   /* post lwidth */

    wi.XOrgX = xpad + strwidth("-000.00", LabelFont);
    wi.XOrgY = ypad + /* LabelHeight */ 8;
    wi.XOppX = wi.area_w - lwidth - /* legend pad */ 13;
    wi.XOppY = wi.area_h - /* TitleSpace */ 50;
    wi.XUnitsPerPixel =	(wi.UsrOppX - wi.UsrOrgX)/(wi.XOppX -wi.XOrgX);
    wi.YUnitsPerPixel =	(wi.UsrOppY - wi.UsrOrgY)/(wi.XOppY -wi.XOrgY);

    if (pmBool("asp")) {
	double aspect = pmDouble("aspect");
	double tmp = wi.XUnitsPerPixel/wi.YUnitsPerPixel;
	if (tmp > aspect) {
	    /* adjust to width */
	    double k = aspect/tmp, XOrgY = wi.XOrgY;
	    wi.XOrgY = (int) (XOrgY*(1 + k)/2 + wi.XOppY*(1 - k)/2);
	    wi.XOppY = (int) (XOrgY*(1 - k)/2 + wi.XOppY*(1 + k)/2);
	    wi.YUnitsPerPixel  = wi.XUnitsPerPixel/aspect;
	} else {
	    /* adjust to height */
	    double k = tmp/aspect, XOrgX = wi.XOrgX;
	    wi.XOrgX = (int) (XOrgX*(1 + k)/2 + wi.XOppX*(1 - k)/2);
	    wi.XOppX = (int) (XOrgX*(1 - k)/2 + wi.XOppX*(1 + k)/2);
	    wi.XUnitsPerPixel = wi.YUnitsPerPixel*aspect;
	}
	pm_set("aspect", aspect);
    } else {
	double aspect = wi.XUnitsPerPixel/wi.YUnitsPerPixel;
	pm_set("aspect", aspect);
    }

    return &wi;
}

/****************************************
 * Grid setup function
 ---------------------------------------*/
double gridBase, gridStep, gridJuke[101];
int gridNJuke, gridCurJuke;

double RoundUp(double val)
{
    int exponent = (int) floor(log10(val));
    val /= pow(10, exponent);
    val = (val > 5.0)? 10.0:
          (val > 2.0)?  5.0:
          (val > 1.0)?  2.0:
          /*  else  */  1.0;
    return (val * pow(10, exponent));
}

double stepGrid(void)
{
    if (++gridCurJuke >= gridNJuke) {
        gridCurJuke = 0;
        gridBase += gridStep;
    }
    return(gridBase + gridJuke[gridCurJuke]);
}

#define ADD_GRID(val)   (gridJuke[gridNJuke++] = log10(val))

double initGrid(double low, double step, int logFlag)
{
    gridNJuke = gridCurJuke = 0;
    gridJuke[gridNJuke++] = 0.0;

    if (logFlag) {
	double ratio, x;
	ratio = pow(10.0, step);
        gridBase = floor(low);
        gridStep = ceil(step);
        if (ratio <= 3.0) {                /* if ratio > 3./1. no gridJuke  */
            if (ratio > 2.0) {             /* 2.0 = 2.0/1.0 = 10.0/5.0      */
                ADD_GRID(3.0);
            } else if (ratio > 2.0/1.5) {  /* 2.0/1.5 < 7/5 < 10/7 < 3/2 ...*/
                ADD_GRID(2.0);  ADD_GRID(5.0);
            } else if (ratio > 2.5/2.0) {  /* next grid must be '2.5'       */
                ADD_GRID(1.5);  ADD_GRID(2.0);  ADD_GRID(3.0);
                ADD_GRID(5.0);  ADD_GRID(7.0);
            } else {
                for (x = 1.0; x < 10.0 && (x+.5)/(x+.4) >= ratio; x += .5) {
                    ADD_GRID(x + .1);   ADD_GRID(x + .2);
                    ADD_GRID(x + .3);   ADD_GRID(x + .4);
                    ADD_GRID(x + .5);
                }
                if (floor(x) != x) ADD_GRID(x += .5);
                for ( ; x < 10.0 && (x+1.0)/(x+.5) >= ratio; x += 1.0) {
                    ADD_GRID(x + .5);   ADD_GRID(x + 1.0);
                }
                for ( ; x < 10.0 && (x+1.0)/x >= ratio; x += 1.0) {
                    ADD_GRID(x + 1.0);
                }
                if (x == 7.0) { gridNJuke--; x = 6.0; }
		if (x <  7.0) { ADD_GRID(x + 2.0); }
                if (x ==10.0) { gridNJuke--; }
            }
            x = low - gridBase;
	    gridCurJuke = -1;
            while (x >= gridJuke[gridCurJuke+1])
		gridCurJuke++;
        }
    } else {
	gridStep = RoundUp(step);
	gridBase = floor(low/gridStep) * gridStep;
    }
    return (stepGrid());
}

/***************************************************
 *  glut call back functions
 --------------------------------------------------*/
void draw(void)
{
    gldraw();
    glutSwapBuffers();
}

void reshape(int w, int h)
{
    glViewport(0, 0, (GLint) w, (GLint) h);
    glutPostRedisplay();
}

void action(int key)
{
    static double aspect = pmDouble("aspect");
    static int
	Axis        = pmBool("Axis"),
	BoundBox    = pmBool("BoundBox"),
	Grid        = pmBool("Grid"),
	LargePixels = pmBool("LargePixels"),
	Markers     = pmBool("Markers"),
	NoLines     = pmBool("NoLines"),
	ReverseVideo= pmBool("ReverseVideo"),
	Ticks       = pmBool("Ticks"),
	asp         = pmBool("asp");

    switch(key) {
    case 'q': exit(0);
    case 'a': Axis = !Axis; break;
    case 'b': BoundBox = !BoundBox; break;
    case 'g': Axis = Grid = !Grid; break;
    case 'G': Axis = Grid = 1; BoundBox = Ticks = 0; break;
    case 'l': NoLines = !NoLines; break;
    case 'L': Markers = LargePixels = NoLines = 0; break;
    case 'm': NoLines = Markers = !Markers; LargePixels = 0; break;
    case 'M': NoLines = Markers = 1; LargePixels = 0; break;
    case 'N': BoundBox = Grid = 0; break;
    case 'p': NoLines = LargePixels = !LargePixels; Markers = 0; break;
    case 'P': NoLines = LargePixels = 1; Markers = 0; break;
    case 'r': ReverseVideo = !ReverseVideo; break;
    case 't': BoundBox = Ticks = !Ticks; Grid = 1; Axis = !Ticks; break;
    case 'T': BoundBox = Grid = Ticks= 1; Axis = 0; break;
    case '.': NoLines = 1; Markers = LargePixels = 0; break;
    case '-': asp = !asp; aspect = 1.0; break;
    case '^': asp = 1; aspect *= 1.2; break;
    case '\\':asp = 1; aspect /= 1.2; break;
    case 's': psdraw(); return;
    case 'B': draw(); glReadBuffer(GL_FRONT);
	      SaveImageAs(pmString("bmpFileName")); return;
    default : break;
    }

    pm_set("Axis", Axis);
    pm_set("BoundBox", BoundBox);
    pm_set("Grid", Grid);
    pm_set("LargePixels", LargePixels);
    pm_set("Markers", Markers);
    pm_set("NoLines", NoLines);
    pm_set("ReverseVideo", ReverseVideo);
    pm_set("Ticks", Ticks);
    pm_set("asp", asp);
    pm_set("aspect", aspect);
    glutPostRedisplay(); return;
}

void keyboard(unsigned char key, int x, int y) { action((int) key); }

/*******************************
 * Mouse
 ------------------------------*/
void pushpop(int pop)
{
    extern double llx, lly, urx, ury;
    const  int STACKSIZE = 32;
    static double bbs[STACKSIZE][4];
    static int sp = 0;

    /* push */
    if (!pop && sp < STACKSIZE) {
	bbs[sp][0] = llx; bbs[sp][1] = lly;
	bbs[sp][2] = urx; bbs[sp][3] = ury;
	sp++; return;
    }
    /* pop */
    if (pop && sp) {
	sp--;
	llx = bbs[sp][0]; lly = bbs[sp][1];
	urx = bbs[sp][2]; ury = bbs[sp][3];
	return;
    }
}
inline void push(void) { pushpop(0); }
inline void pop(void)  { pushpop(1); }

void rectXor(double x0, double y0, double x1, double y1)
{
    glEnable(GL_COLOR_LOGIC_OP);
    glLogicOp(GL_INVERT);
    glBegin(GL_LINE_LOOP); {
	glVertex2d(x0, y0);
	glVertex2d(x1, y0);
	glVertex2d(x1, y1);
	glVertex2d(x0, y1);
    } glEnd();
    glDisable(GL_COLOR_LOGIC_OP);
}

#define GLUT_MOVE  2    /* GLUT_DOWN = 0, GLUT_UP = 1 */

void mouse(int button, int state, int x, int y)
{
    extern double llx, lly, urx, ury;
    static Point2d pt, pt0;
    static int x0, y0;
    static GLdouble mvmatrix[16], projmatrix[16], dummy;
    static GLint viewport[4];
    static int selecting = 0;

    if (button != GLUT_LEFT_BUTTON) return;
    switch (state) {
    case GLUT_DOWN:
	glGetDoublev(GL_MODELVIEW_MATRIX, mvmatrix);
	glGetDoublev(GL_PROJECTION_MATRIX, projmatrix);
	glGetIntegerv(GL_VIEWPORT, viewport);
	y = viewport[3] - 1 - y;  /* mouse pos to raster pos */
	gluUnProject(x, y, 0.0, mvmatrix, projmatrix, viewport,
		     &pt.x, &pt.y, &dummy);
	pt0 = pt;
	x0 = x; y0 = y;
	selecting = 1;
	glDrawBuffer(GL_FRONT);
	return;
    case GLUT_MOVE:
	if (selecting == 0) return;
	y = viewport[3] - 1 - y;
	rectXor(pt0.x, pt0.y, pt.x, pt.y);
	gluUnProject(x, y, 0.0, mvmatrix, projmatrix, viewport,
		     &pt.x, &pt.y, &dummy);
	rectXor(pt0.x, pt0.y, pt.x, pt.y);
	glFlush();
	return;
    case GLUT_UP:
	if (selecting == 0) return;
	y = viewport[3] - 1 - y;
	if (abs(x - x0) + abs(y - y0) > 10) {  /* new regeon threshold */
	    push();
	    llx = MIN(pt0.x, pt.x); lly = MIN(pt0.y, pt.y);
	    urx = MAX(pt0.x, pt.x); ury = MAX(pt0.y, pt.y);
	} else {
	    pop();                     /* return to the previous scene */
	}
	selecting = 0;
	glDrawBuffer(GL_BACK);
	glutPostRedisplay();
	return;
    default:
	return;
    }
}

void motion(int x, int y) { mouse(GLUT_LEFT_BUTTON, GLUT_MOVE, x, y); }

/*******************************
 * Menu setup
 ------------------------------*/
void make_menu(void)
{
    char PSOutTo[128];
    char BMPOutTo[128];
    sprintf(PSOutTo, "PostScript (%s)", pmString("psFileName"));
    sprintf(BMPOutTo, "SaveImage (%s)", pmString("bmpFileName"));

    int GridStyle = glutCreateMenu(action); {
        glutAddMenuEntry("-FullGrid",    'G');
        glutAddMenuEntry("-Ticks",       'T');
        glutAddMenuEntry("-None",        'N');
	glutAddMenuEntry("Toggle BBox",  'b');
	glutAddMenuEntry("Toggle Axis",  'a');
    }
    int LineStyle = glutCreateMenu(action); {
        glutAddMenuEntry("Normal Lines", 'L');
        glutAddMenuEntry("Markers",      'M');
        glutAddMenuEntry("Large Pixels", 'P');
        glutAddMenuEntry("Small Pixels", '.');
        glutAddMenuEntry("Toggle Lines", 'l');
    }
    glutCreateMenu(action); {
	glutAddMenuEntry("Quit",              'q');
	glutAddSubMenu("GridStyle",     GridStyle);
	glutAddSubMenu("LineStyle",     LineStyle);
	glutAddMenuEntry("Toggle Auto Scale", '-');
	glutAddMenuEntry("Reverse Video",     'r');
	glutAddMenuEntry( PSOutTo,            's');
	glutAddMenuEntry( BMPOutTo,           'B');
    }
    glutAttachMenu(GLUT_RIGHT_BUTTON);
}

/***************************************************
 * Main 
 --------------------------------------------------*/
int main(int argc, char *argv[])
{
    void initParams(void);
    int  ParseArgs(int argc, char *argv[], char *FileArray[] = NULL);
    void ReadData(char *FileNames);
    void Traverse(int logXFlag, int logYFlag, 
		  double *loX, double *hiX, double *loY, double *hiY);
    int   numFiles = 0;              /* Number of input files   */
    char *inFileNames[MAXSETS];      /* File names              */
    
    /* Init data sets */
    for (int i = 0; i < MAXSETS; i++) {
        PlotData[i].setName = NULL;
        PlotData[i].list = (PointList *) NULL;
    }
    initParams();

    /* read options and input file names */
    numFiles = ParseArgs(argc, argv, inFileNames);

    /* read data */
    if (numFiles > 0) {
	char psFileName[128], bmpFileName[128];
	strcpy(psFileName, inFileNames[0]);
	strtok(psFileName, ".");   strcpy(bmpFileName, psFileName);
	strcat(psFileName, ".ps"); strcat(bmpFileName, ".bmp");
	pm_set("psFileName", psFileName);
	pm_set("bmpFileName", bmpFileName);
	for (int i = 0; i < numFiles; i++)
	    ReadData(inFileNames[i]);
    } else {
	ReadData("stdin");
    }

    ParseArgs(argc, argv); // command line option should overwrite file's one

    /* read in user bounding box */
    Traverse(LogX(), LogY(), &llx, &urx, &lly, &ury);
    double loX = pmDouble("XLowLimit"), hiX = pmDouble("XHighLimit"),
	   loY = pmDouble("YLowLimit"), hiY = pmDouble("YHighLimit");
    if (loX < hiX) llx = loX, urx = hiX;
    if (loY < hiY) lly = loY, ury = hiY;

    /* Postscript */ 
    if (pmBool("ps")) { psdraw(); exit(0); }
    
    /* init glut window */
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
    glutInitWindowSize(pmInt("Width"), pmInt("Height"));
    glutCreateWindow(argv[0]);
    
    /* other init: pixel packing mode */
    glPixelStorei(GL_PACK_ALIGNMENT, 1);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

    /* call back functions */
    glutDisplayFunc(draw);
    glutReshapeFunc(reshape);
    glutKeyboardFunc(keyboard);
    glutMotionFunc(motion);
    glutMouseFunc(mouse);
    make_menu();

    glutMainLoop();
    return 0;
}

/*****************************************
 * read default parameter values
 ----------------------------------------*/
void initParams(void)
{
    /* strings */
    pm_set("TitleText",  "glGraph");
    pm_set("psFileName", "glgraph.ps");
    pm_set("bmpFileName", "glgraph.bmp");
    pm_set("XUnitText", "X");
    pm_set("YUnitText", "Y");
    /* fonts */
    pm_set("LabelFont", "helvetica-12");
    pm_set("TitleFont", "helvetica-18");
    pm_set("ps.LabelFont", "Times-Roman");
    pm_set("ps.TitleFont", "Times-Bold");
    /* values */
    pm_set("Width",       WIDTH);
    pm_set("Height",      HEIGHT);
    pm_set("XLowLimit",  "0.0");  /* Low <= High means defined from data */
    pm_set("XHighLimit", "0.0");
    pm_set("YLowLimit",  "0.0");
    pm_set("YHighLimit", "0.0");
    /* flags */
    pm_set("LogX",        "off");
    pm_set("LogY",        "off");
    pm_set("Axis",         "on");
    pm_set("Grid",         "on");
    pm_set("BoundBox",    "off");
    pm_set("LargePixels", "off");
    pm_set("Markers",     "off");
    pm_set("NoLines",     "off");
    pm_set("PixelMarkers","off");
    pm_set("ReverseVideo","off");
    pm_set("Ticks",       "off");
    pm_set("asp",         "off");
    pm_set("ps",          "off");
    /* colors */
    pm_set("Background", "#c0c0c0");
    pm_set("Foreground", "#000000");
    pm_set("ZeroColor",  "#ffffff");
    pm_set("0.Color",    "#ff0000"); // red         {255,  0,  0};
    pm_set("1.Color",    "#00ff7f"); // SpringGreen {  0,255,127};
    pm_set("2.Color",    "#0000ff"); // blue        {  0,  0,255};
    pm_set("3.Color",    "#ffff00"); // yellow      {255,255,  0};
    pm_set("4.Color",    "#00ffff"); // cyan        {  0,255,255};
    pm_set("5.Color",    "#a0522d"); // sienna      {160, 82, 45};
    pm_set("6.Color",    "#ffa500"); // orange      {255,165,  0};
    pm_set("7.Color",    "#ff7f50"); // coral       {255,127, 80};
}

/*****************************************
 * ParseArgs 
 ----------------------------------------*/
void usage(void)
{
    fprintf(stderr, "\n");
    fprintf(stderr,
    "glgraph [-<digit> set_name] [-asp aspect] [-bg background_color] [-bb]\n"
    "        [-fg foreground_color] [-lf label_font] [-lnx] [-lny]\n"
    "        [-lx xlow,xhigh] [-ly ylow,yhigh] [-ng] [-nl] [-m] [-P]"
                                                     " [-ps] [-rv]\n"
    "        [-t title] [-tf title_font] [-tk] [-x unit_name] [-y unit_name]\n"
    "        [-zg zero_color] [=WxH] file...\n"
	    );
    fprintf(stderr, "\n");
    fprintf(stderr, "-bb   Draw bounding box around data\n");
    fprintf(stderr, "-lnx  Logarithmic scale for X axis\n");
    fprintf(stderr, "-lny  Logarithmic scale for Y axis\n");
    fprintf(stderr, "-m    Mark points distinctively\n");
    fprintf(stderr, "-ng   Don't draw grid or tick marks\n");
    fprintf(stderr, "-nl   Don't draw lines(scattered plot)\n");
    fprintf(stderr, "-P    Mark points with dot\n");
    fprintf(stderr, "-ps   Act as a PostScript converter\n");
    fprintf(stderr, "-rv   Reverse Video\n");
    fprintf(stderr, "-tk   Draw tick marks instead of full grid\n");
    fprintf(stderr, "\n");
}

void eprintf(char *fmt, ...) 
{
    va_list ap; va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    exit(1);
}

#define ARG(opt, name) \
if (strcmp(argv[idx], opt) == 0) { \
    if (!inFileNames) { pm_set(#name, "on"); } \
    continue; \
}

#define ARG1(opt, name, message) \
if (strcmp(argv[idx], opt) == 0) { \
    if (++idx >= argc) \
        eprintf("Error in %s: %s\n\n", argv[--idx], message); \
    if (!inFileNames) { pm_set(#name, argv[idx]); }\
    continue; \
}

#define ARG2(opt, name1, name2, Type, fmt, message) \
if (strcmp(argv[idx], opt) == 0) { \
    Type name1, name2; \
    if (++idx >= argc || sscanf(argv[idx], fmt, &name1, &name2) != 2) \
        eprintf("Error in %s: %s\n\n", argv[--idx], message); \
    if (!inFileNames) { pm_set(#name1, name1); pm_set(#name2, name2); }\
    continue; \
}

int ParseArgs(int argc, char *argv[], char *inFileNames[])
{
    int set, numFiles = 0;

    for (int idx = 1; idx < argc; idx++) {
	if (*argv[idx] == '-') {
            if (sscanf(argv[idx], "-%d", &set) == 1) { /* set name */
                if (++idx >= argc)
		    eprintf("Error in %s: %s\n\n",
			    argv[--idx],"missing set name");
                if (!inFileNames) { PlotData[set].setName = argv[idx]; }
		continue;
            }
	    ARG("-bb",  BoundBox);
	    ARG("-db",  debug);
	    ARG("-lnx", LogX);
	    ARG("-lny", LogY);
	    ARG("-m",   Markers);
	    ARG("-nl",  NoLines);
	    ARG("-p",   PixelMarkers);
	    ARG("-P",   LargePixels);
	    ARG("-rv",  ReverseVideo);
	    ARG("-tk",  Ticks);
	    ARG1("-bg", Background, "missing color");
	    ARG1("-fg", Foreground, "missing color");
	    ARG1("-lf", LabelFont, "missing font name");
	    ARG1("-tf", TitleFont, "missing font name");
	    ARG1("-t",  TitleText, "missing plot title");
	    ARG1("-x",  XUnitText, "missing axis name");
	    ARG1("-y",  YUnitText, "missing axis name");
	    ARG1("-zg", ZeroColor, "missing color");
	    ARG2("-lx", XLowLimit, XHighLimit, double, "%lf,%lf",
		  "format must be -lx <x-low>,<x-high>");
	    ARG2("-ly", YLowLimit, YHighLimit, double, "%lf,%lf",
		 "format must be -ly <y-low>,<y-high>");

	    /* original options */
	    ARG("-ps",ps);
	    ARG1("-ps.lf", ps.LabelFont, "missing font name");
	    ARG1("-ps.tf", ps.TitleFont, "missing font name");
	    if (strcmp(argv[idx], "-ng") == 0) { 
		if (!inFileNames) pm_set("Grid", "off");
		continue;
	    }
	    if (strcmp(argv[idx], "-asp") == 0) {
		double aspect;
		if (sscanf(argv[++idx], "%lf", &aspect) != 1)
		    eprintf("%s...: missing aspect ratio\n", argv[--idx]);
		if (!inFileNames) {
		    pm_set("asp", "on"); pm_set("aspect", aspect);
		}
		continue;
	    }
	    ARG2("-g", Width, Height, int, "%dx%d",
		 "format must be -g <width>x<height>");

	    /* else */
	    fprintf(stderr, "unknown option: %s\n", argv[idx]);
	    usage(); exit(1);

	} else if (*argv[idx] == '=') {
	    int width, height;
	    if (sscanf(argv[idx]+1,"%dx%d", &width, &height) != 2)
		eprintf("format must be \"=<width>x<height>\"\n");
	    if (!inFileNames) {
		pm_set("Width", width); pm_set("Height", height);
	    }
	    continue;
        } else if (*argv[idx] != '&') { /* Should be an input file */
	    if (inFileNames)
		inFileNames[numFiles] = argv[idx];
	    numFiles++;
	    continue;
        }
    }
    if (!inFileNames && pmBool("Ticks")) pm_set("Axis", "off");
    return numFiles;
}

/****************************************
 * read data
 ---------------------------------------*/
#define STRDUP(S)  (strcpy((char *)malloc((unsigned) (strlen(S)+1)), (S)))

/*
 * if redundant set change the set name 
 * else prepare a new set and return entry point for lists 
 */
PointList **NewSet(char *setName)
{
    static int numSets = 0;

    if (numSets > 0 &&  PlotData[numSets-1].list &&
	PlotData[numSets-1].list->numPoints == 0)   /* redundant set */
    {
	if (PlotData[numSets-1].setName) free(PlotData[numSets-1].setName);
	PlotData[numSets-1].setName = STRDUP(setName);
	return &(PlotData[numSets-1].list);
    } else if (++numSets <= MAXSETS) {
	if (PlotData[numSets-1].setName) free(PlotData[numSets-1].setName);
	PlotData[numSets-1].setName = STRDUP(setName);
	PlotData[numSets-1].list = NULL;
	return &(PlotData[numSets-1].list);
    } else {
	return NULL;
    }
}

PointList *NewList(void)
{
    const int INITSIZE = 128;

    PointList *list = (PointList *) malloc(sizeof(PointList));
    list->pt = (Point2d *) malloc(sizeof(Point2d)*INITSIZE);
    list->numPoints = 0;
    list->allocSize = INITSIZE;
    list->next      = NULL;
    return list;
}

void EndList(PointList *list)
{
    if (list && list->numPoints != list->allocSize) { 
	Point2d *pt = (Point2d *) realloc(list->pt,
					  sizeof(Point2d)*list->numPoints);
	if (pt) {
	    list->pt = pt;
	    list->allocSize = list->numPoints;
	}
    } 
}

void AddPoint(PointList *list, Point2d *pt)
{
    if (!list) { fprintf(stderr, "Bad list"); exit(1); }
    if (list->numPoints >= list->allocSize) {
	list->allocSize *=2;
	list->pt = (Point2d *) realloc(list->pt,
				       sizeof(Point2d)*list->allocSize);
    }
    list->pt[list->numPoints++] = *pt;
}

void ReadData(char *fname)
{
    static PointList **next   = NULL;
    static PointList *curlist = NULL;
    static char linebuf[256], setName[128];
    static int idx;
    FILE *fp;

    if (strcmp(fname, "stdin") == 0) {
	fp = stdin;
	sprintf(setName, "Set %d", idx);
    } else {
	if (!(fp = fopen(fname, "r"))) {
	    fprintf(stderr, "\ncannot open file '%s'\n\n", fname); exit(1);
	}
	char *c = &fname[strlen(fname)-1];
	while (c > fname && *(c-1) != '\\' && *(c-1) != '/') c--;
	sprintf(setName, "%s", c);
    }

    /* new set */
    next = NewSet(setName);
    curlist = *next = NewList();
    next = &curlist->next;;
    idx++;
    while (fgets(linebuf, sizeof(linebuf), fp)) {
	static int newset = 0;  // state flag 
	Point2d pt;
	char *line = linebuf, *pname, *pval;
	while (*line && isspace(*line)) line++;
	if (*line =='\0') {                      /* empty line */
	    if (curlist && curlist->numPoints > 0) {
		EndList(curlist);
		newset = 1;
		if (fp == stdin)
		    sprintf(setName, "Set %d", idx);
	    }
	    continue;
	} else {
	    if (*line == '#') { continue; }      /* comment */
	    if (*line == '"') {                  /* set name */
		char *sname = setName;
		line++;
		while (*line && *line != '\n' && *line != '"')
		    *sname++ = *line++;
		*sname = '\0';
		if (curlist && curlist->numPoints > 0)
		    EndList(curlist);
		newset = 1; 
		continue;
	    }
	    /* point data */
	    if (sscanf(line, "move %lf %lf", &pt.x, &pt.y) == 2) { /* move */
		if (curlist)
		    if (curlist->numPoints == 0) {
			AddPoint(curlist, &pt); continue;
		    } else
			EndList(curlist);
		curlist = *next = NewList();
		next = &curlist->next;
		AddPoint(curlist, &pt);
		continue;
	    }
	    if (sscanf(line, "%lf %lf", &pt.x, &pt.y) == 2 ||
		sscanf(line, "draw %lf %lf", &pt.x, &pt.y) == 2) { /* draw */
		if (newset) {
		    next = NewSet(setName);
		    curlist = *next = NewList();
		    next = &curlist->next;;
		    idx++; newset = 0;
		}
		AddPoint(curlist, &pt);
		continue;
	    }
	    /* options */
	    pname = strtok(line, " ");
	    if (pname[strlen(pname) - 1] == ':') {     /* parameter name */
		pname[strlen(pname) - 1] = '\0';
		pval = strtok(NULL, "\n");
		while (*pval && isspace(*pval)) pval++;
		pm_set(pname, pval);
		continue;
	    }
	}
    }
    if (curlist) EndList(curlist);
    if (pmBool("Ticks")) pm_set("Axis", "off");
    fclose(fp);
}

void Traverse(int logXFlag, int logYFlag, 
	      double *loX, double *hiX, double *loY, double *hiY)
{
    double 
	xmin = FLT_MAX, xmax = -FLT_MAX,
	ymin = FLT_MAX, ymax = -FLT_MAX;

    for (int i = 0; i < MAXSETS; i++) {
	PointList *list = PlotData[i].list;
	for ( ; list; list = list->next) {
	    for (int j = 0; j < list->numPoints; j++) {
		if (logXFlag)
		    if (list->pt[j].x > 0.0) {
			list->pt[j].x = log10(list->pt[j].x);
		    } else {
			fprintf(stderr, "non-positive X-data\n"); exit(1);
		    }
		if (logYFlag)
		    if (list->pt[j].y > 0.0) {
			list->pt[j].y = log10(list->pt[j].y);
		    } else {
			fprintf(stderr, "non-positive Y-data\n"); exit(1);
		    }
		xmin = MIN(xmin, list->pt[j].x);
		xmax = MAX(xmax, list->pt[j].x);
		ymin = MIN(ymin, list->pt[j].y);
		ymax = MAX(ymax, list->pt[j].y);
	    }
	}
    }
    if (loX) *loX = xmin;
    if (hiX) *hiX = xmax;
    if (loY) *loY = ymin;
    if (hiY) *hiY = ymax;
}
