/***********************************************************************
 * glgraph.h
 ----------------------------------------------------------------------*/
/* PointList */
typedef struct point2d {
    double x, y;
} Point2d;

typedef struct point_list {
    int numPoints;
    int allocSize;
    Point2d *pt;
    struct point_list *next;
} PointList;

typedef struct data_set {
    char *setName;
    PointList *list;
} DataSet;

#define MAXATTR         8       /* num Color/LineStyle etc. */
#define MAXSETS        64

/* Main object of glGraph */
extern DataSet PlotData[MAXSETS];

/* Basic transformation stuff */
typedef struct _localwin {
    int    area_w, area_h;      /* Width and height of screen in pixels */
    int    XOrgX, XOrgY;	/* Origin of bounding box on screen     */
    int    XOppX, XOppY;	/* Other point defining bouding box     */
    double UsrOrgX, UsrOrgY;	/* Origin of bounding box in user space */
    double UsrOppX, UsrOppY;	/* Other point of bouding box           */
    double XUnitsPerPixel;	/* X Axis scale factor                  */
    double YUnitsPerPixel;	/* Y Axis scale factor                  */
    //void   *dev_info;		/* Device information                   */ 
} LocalWin;

extern LocalWin *getWinInfo(void); /* return Info of the current window */

/* user coordinate to window raster position */
#define SCREENX(x)  (int(((x) - wi->UsrOrgX)/wi->XUnitsPerPixel + wi->XOrgX))
#define SCREENY(y)  (int(((y) - wi->UsrOrgY)/wi->YUnitsPerPixel + wi->XOrgY))

/*****************************************
 * Text justifications
 ----------------------------------------*/
#define T_CENTER       0
#define T_LEFT         1
#define T_UPPERLEFT    2
#define T_TOP          3
#define T_UPPERRIGHT   4
#define T_RIGHT        5
#define T_LOWERRIGHT   6
#define T_BOTTOM       7
#define T_LOWERLEFT    8

/*****************************************
 * glgraph.c: grid setup
 ----------------------------------------*/
extern double initGrid(double low, double step, int logFlag);
extern double stepGrid(void);

/*****************************************
 * params.c: parameter table
 ----------------------------------------*/
extern int    pmBool(const char *Key);
extern int    pmInt(const char *Key);
extern double pmDouble(const char *Key);
extern char  *pmString(const char *Key);
extern float *pmColor(const char *Key);

#define LogX()         pmBool("LogX")
#define LogY()         pmBool("LogY")
#define Axis()         pmBool("Axis")
#define BoundBox()     pmBool("BoundBox")
#define PixelMarkers() pmBool("PixelMarkers")
#define Grid()         pmBool("Grid")
#define LargePixels()  pmBool("LargePixels")
#define Markers()      pmBool("Markers")
#define NoLines()      pmBool("NoLines")
#define Ticks()        pmBool("Ticks")
#define ReverseVideo() pmBool("ReverseVideo")

/*****************************************
 * draw functions
 ----------------------------------------*/
extern void gldraw(void);
extern void psdraw(void);
