/*****************************************
 * params.c
 ----------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string>
#include <map>
#include "glgraph.h"

typedef std::basic_string<char> string;

std::map<std::string, char *> pm;

#define STRDUP(S)  (strcpy((char *)malloc((unsigned) (strlen(S)+1)), (S)))

char *lower(const char *str) {
    static char s[100];
    for (int i = 0; (s[i] = tolower(str[i])); i++);
    return s;
}

int stricmp(char *a, char *b, int n = 256)
{
    int value = 0;
    if ((a == (char *) 0) || (b == (char *) 0)) { return a - b; }
    for (int i = 0;
         (i < n && (*a | *b) &&
          !(value = ((isupper(*a) ? *a - 'A' + 'a' : *a) -
                     (isupper(*b) ? *b - 'A' + 'a' : *b))));
         a++, b++, i++)
        /* empty body */ ;
    return value;
}

void pm_set(const char *Key, char *val)
{
    string key = string(lower(Key));
    if (pm[key]) free(pm[key]);
    pm[key] = STRDUP(val);
}

void pm_set(const char *Key, double val)
{
    string key = string(lower(Key));
    char str[100]; sprintf(str, "%f", val);
    if (pm[key]) free(pm[key]);
    pm[key] = STRDUP(str);
 }

void pm_set(const char *Key, int val)
{
    string key = string(lower(Key));
    char str[100]; sprintf(str, "%d", val);
    if (pm[key]) free(pm[key]);
    pm[key] = STRDUP(str);
}

double pmDouble(const char *Key) 
{
    string key = string(lower(Key));
    return (pm[key])? atof(pm[key]) : 0;
}

int pmInt(const char *Key)
{
    string key = string(lower(Key));
    return (pm[key])? atoi(pm[key]) : 0;
}

char *pmString(const char *Key)
{
    string key = string(lower(Key));
    return (pm[key])? pm[key] : NULL;
}

float *pmColor(const char *Key)
{
    string key = string(lower(Key));
    if (pm[key]) {
	if (*pm[key] == '#') {               /* first time scan */
	    float color[3] = {.0,.0,.0};
	    char *str = pm[key] + 1;
	    int  cbit = (strlen(str)/3)*4;
	    for (int i = 0; i < 3; i++, str += cbit/4) {
		unsigned short val;
		(cbit == 4) ? sscanf(str, "%1hx", &val):
		(cbit == 8) ? sscanf(str, "%2hx", &val):
		 /* else */   sscanf(str, "%4hx", &val);
		color[i] = float(val)/(1<<cbit);
	    }
	    free(pm[key]); pm[key] = (char *)malloc(sizeof(float)*3);
	    memcpy(pm[key], color, sizeof(float)*3);
	}
	return (float *)pm[key];
    } else {
	return NULL;
    }
}

int pmBool(const char *Key)   
{
    static char *positive[] = { "1", "on",  "true",  "yes", (char *)0 };
    static char *negative[] = { "0", "off", "false", "no",  (char *)0 };
    string key = string(lower(Key));
    char **term;

    for (term = positive; *term; term++)
	if (stricmp(pm[key], *term) == 0) break;
    if (*term) return 1;
    for (term = negative; *term; term++)
	if (stricmp(pm[key], *term) == 0) break;
    if (*term) return 0;
    /* else */
    fprintf(stderr, "Error in pmBool: pm[%s] = \"%s\"\n", Key, pm[key]);
    return 0;

}
