/***************************************************
 * psdraw.c: PostScript output
 --------------------------------------------------*/
#include <stdio.h>
#include <math.h>
#include <string>
#include "glgraph.h"

#define MAX(a,b)     ((a > b) ? a : b)
#define MIN(a,b)     ((a < b) ? a : b)

/*
 * Macros for drawing title and axis labels
 */
void psFonts(FILE *fp)
{
    fprintf(fp,
	    "/choose-font    %% stack: fontsize fontname => ---\n"
	    "{\n"
	    "  findfont exch scalefont setfont\n"
	    "  gsave\n"
	    "  newpath 0 0 moveto (0) true charpath flattenpath pathbbox\n"
	    "  /top exch def pop  /bottom exch def pop\n"
	    "  bottom top bottom top add 2 div\n"
	    "  /center-font-val exch def\n"
	    "  /upper-font-val  exch def\n"
	    "  /lower-font-val  exch def\n"
	    "  grestore\n"
	    "} def\n"
	    "\n"
	    "%%%% ***-just    %% stack: x y => newx newy\n"
	    "/center-x-just {exch 2 index stringwidth pop 2 div sub exch}"
	    " def\n"
	    "/left-x-just   {} def\n"
	    "/right-x-just  {exch 2 index stringwidth pop sub exch} def\n"
	    "/center-y-just {center-font-val sub} def\n"
	    "/lower-y-just  {lower-font-val  sub} def\n"
	    "/upper-y-just  {upper-font-val  sub} def\n"
	    "\n"
	    "/just-string   %% stack: (string) x y just => ----\n"
	    "{\n"
	    "  dup 0 eq {pop aspect mul center-x-just center-y-just} if\n"
	    "  dup 1 eq {pop aspect mul left-x-just   center-y-just} if\n"
	    "  dup 2 eq {pop aspect mul left-x-just   upper-y-just } if\n"
	    "  dup 3 eq {pop aspect mul center-x-just upper-y-just } if\n"
	    "  dup 4 eq {pop aspect mul right-x-just  upper-y-just } if\n"
	    "  dup 5 eq {pop aspect mul right-x-just  center-y-just} if\n"
	    "  dup 6 eq {pop aspect mul right-x-just  lower-y-just } if\n"
	    "  dup 7 eq {pop aspect mul center-x-just lower-y-just } if\n"
	    "  dup 8 eq {pop aspect mul left-x-just   lower-y-just } if\n"
	    "  moveto show\n"
	    "} def\n"
	    "%%\n"
	    "  /ul {2} def   /t {3} def   /ur {4} def\n"
	    "  /l  {1} def   /c {0} def   /r  {5} def\n"
	    "  /ll {8} def   /b {7} def   /lr {6} def\n"
	    "\n"
	    );
}

/*
 * Line styles
 */
void psLines(FILE *fp)
{
    fprintf(fp,
	    "%% Line style definition\n"
	    "/solid {[ ] 0 mm setdash} bind def\n"
	    "/dash0 {[ ] 0 mm setdash} bind def\n"
	    "/dash1 {[0.70 mm 0.40 mm] 0 mm setdash} bind def\n"
	    "/dash2 {[1.00 mm 0.50 mm] 0 mm setdash} bind def\n"
	    "/dash3 {[1.50 mm 0.75 mm] 0 mm setdash} bind def\n"
	    "/dash4 {[2.00 mm 1.00 mm] 0 mm setdash} bind def\n"
	    "/dash5 {[2.70 mm 1.40 mm] 0 mm setdash} bind def\n"
	    "/dash6 {[3.50 mm 1.80 mm] 0 mm setdash} bind def\n"
	    "/dash7 {[4.50 mm 2.30 mm] 0 mm setdash} bind def\n"
	    "%%\n"
	    "/dashA {[ 2 mm .5 mm] 1 mm setdash} bind def\n"
	    "/dashB {[10 mm .5 mm .8 mm .5 mm] 3 mm setdash} bind def\n"
	    "/dashC {[13 mm .4 mm .8 mm .4 mm .8 mm .4 mm] 5 mm setdash}"
	    " bind def\n"
	    "\n"
	    );
}

/*
 * Marker definitions
 */
void psMarks(FILE *fp)
{
    fprintf(fp,
	    "%% Marker definition\n"
	    "/msize {.5 mm} def\n"
	    "/mark0 {currentpoint /y exch def /x exch def gsave newpath\n"
	    "  x msize sub y msize sub moveto\n"
	    "  x msize add y msize sub lineto\n"
	    "  x msize add y msize add lineto\n"
	    "  x msize sub y msize add lineto\n"
	    "  closepath fill\n"
	    "  grestore} def\n"
	    "%%\n"
	    "/mark1 {currentpoint /y exch def /x exch def gsave newpath\n"
	    "  x msize sub y msize sub moveto\n"
	    "  x msize add y msize sub lineto\n"
	    "  x msize add y msize add lineto\n"
	    "  x msize sub y msize add lineto\n"
	    "  closepath solid .15 mm setlinewidth stroke\n"
	    "  grestore} def\n"
	    "%%\n"
	    "/mark2 {currentpoint /y exch def /x exch def gsave newpath\n"
	    "  x y msize 0 360 arc\n"
	    "  closepath solid .15 mm setlinewidth stroke\n"
	    "  grestore} def\n"
	    "%%\n"
	    "/mark3 {currentpoint /y exch def /x exch def gsave newpath\n"
	    "  x msize sub y msize sub moveto\n"
	    "  x msize add y msize add lineto\n"
	    "  x msize sub y msize add moveto\n"
	    "  x msize add y msize sub lineto\n"
	    "  solid .15 mm setlinewidth stroke\n"
	    "  grestore} def\n"
	    "%%\n"
	    "/mark4 {currentpoint /y exch def /x exch def gsave newpath\n"
	    "  x msize sub y moveto\n"
	    "  x y msize sub lineto\n"
	    "  x msize add y lineto\n"
	    "  x y msize add lineto\n"
	    "  closepath solid .15 mm setlinewidth stroke\n"
	    "  grestore} def\n"
	    "%%\n"
	    "/mark5 { mark1\n"
	    " currentpoint /y exch def /x exch def gsave newpath\n"
	    "  x msize sub y moveto\n"
	    "  x msize add y lineto\n"
	    "  solid .15 mm setlinewidth stroke\n"
	    "  grestore} def\n"
	    "%%\n"
	    "/mark6 { currentpoint /y exch def /x exch def gsave newpath\n"
	    "  x y msize 0 360 arc\n"
	    "  closepath fill\n"
	    "  grestore } def\n"
	    "%%\n"
	    "/mark7 {currentpoint /y exch def /x exch def gsave newpath\n"
	    "  x y moveto\n"
	    "  x msize sub y msize sub lineto\n"
	    "  x msize add y msize sub lineto\n"
	    "  x msize sub y msize add lineto\n"
	    "  x msize add y msize add lineto\n"
	    "  closepath fill\n"
	    "  grestore} def\n"
	    "\n"
	    );
}

void psdraw(void)
{
    LocalWin *wi = getWinInfo();
    char *fname = pmString("psFileName");
    char *TitleText = pmString("TitleText");
    char *XUnitText = pmString("XUnitText");
    char *YUnitText = pmString("YUnitText");
    char *psLabelFont = pmString("ps.LabelFont"); /* default: "Times-Roman" */
    char *psTitleFont = pmString("ps.TitleFont"); /* default: "Times-Bold"  */

    FILE *fp;
    int
	expX = (int)floor(log10(MAX(fabs(wi->UsrOrgX),fabs(wi->UsrOppX)))/3)*3,
	expY = (int)floor(log10(MAX(fabs(wi->UsrOrgY),fabs(wi->UsrOppY)))/3)*3;
    double
	shiftX = 1.0/pow(10, expX),
	shiftY = 1.0/pow(10, expY),
	dX = wi->XUnitsPerPixel*shiftX,
	dY = wi->YUnitsPerPixel*shiftY,
	X0 = wi->UsrOrgX*shiftX, X1 = wi->UsrOppX*shiftX,
	Y0 = wi->UsrOrgY*shiftY, Y1 = wi->UsrOppY*shiftY,
	Left   = X0 - wi->XOrgX*dX,
	Right  = X1 + (wi->area_w  - wi->XOppX)*dX,
        Bottom = Y0 - wi->XOrgY*dY,
	Top    = Y1 + (wi->area_h - wi->XOppY)*dY;
    int  /* PostScript Bounding Box parameters */
	psBBoxW = wi->area_w*4/5,
	psBBoxH = wi->area_h*4/5,
	HOFFSET = 20,
	VOFFSET = 50;

    /* open ps out stream */
    if (!fname || strcmp(fname, "stdout") == 0) {
	fp = stdout; 
    } else {
	if (!(fp = fopen(fname, "w"))) {
	    fprintf(stderr, "\ncannot open file '%s'\n\n", fname); exit(1);
	}
	fprintf(stderr, "\nPostScript out to '%s'\n\n", fname);
    }
    
    fprintf(fp, "%%!PS-Adobe-2.0\n");
    fprintf(fp, "%%%%BoundingBox: %d %d %d %d\n",
	    HOFFSET, VOFFSET, psBBoxW + HOFFSET, psBBoxH + VOFFSET);
    fprintf(fp, "%d %d translate\n", HOFFSET, VOFFSET);
    fprintf(fp, "/BBOX_W %4d def\n", psBBoxW);
    fprintf(fp, "/BBOX_H %4d def\n", psBBoxH);
    fprintf(fp, "%%\n");
    fprintf(fp, "/xscale {BBOX_W %9.4f div} bind def\n", Right - Left);
    fprintf(fp, "/yscale {BBOX_H %9.4f div} bind def\n", Top - Bottom);
    fprintf(fp, "/aspect {yscale xscale div} bind def\n");
    fprintf(fp, "\n");
    fprintf(fp, "/mm{2.8346457 mul xscale div} bind def\n");
    fprintf(fp, "/pt{xscale div} bind def\n");
    fprintf(fp, "/T {aspect mul translate} bind def\n");
    fprintf(fp, "/M {aspect mul moveto} bind def\n");
    fprintf(fp, "/L {aspect mul lineto} bind def\n");
    fprintf(fp, "/p {currentpoint /y exch def /x exch def gsave newpath\n");
    fprintf(fp, "    x y .2 mm 0 360 arc closepath fill grestore} def\n");
    fprintf(fp, "/P {currentpoint /y exch def /x exch def gsave newpath\n");
    fprintf(fp, "    x y .5 mm 0 360 arc closepath fill grestore} def\n");
    fprintf(fp, "\n");
 
    psFonts(fp);
    psLines(fp);
    if (Markers()) psMarks(fp);

    fprintf(fp, "%%%% Scaling\n");
    fprintf(fp, "xscale dup scale\n");
    fprintf(fp, "\n");
    fprintf(fp, "%%%% set Origin\n");
    fprintf(fp, "%9.4f %9.4f T\n", -Left, -Bottom);
    fprintf(fp, "\n");

    if (Grid()) {
	fprintf(fp, "%%%% Title\n");
	fprintf(fp, "5.5 mm /%s choose-font\n", psTitleFont);
	fprintf(fp, "(%s) %9.4f %9.4f  %d just-string\n",
		TitleText, (X0 + X1)/2, Y1 + 24*dY, T_BOTTOM);
	fprintf(fp, "\n");
    }

    if (Axis()) {
	if (!LogY() && Y0 < .0 && Y1 > .0) {
	    fprintf(fp, "%%%% X-axis\n");
	    fprintf(fp, "%9.4f %9.4f M\n", X0, .0);
	    fprintf(fp, "%9.4f %9.4f L\n", X1, .0);
	    fprintf(fp, "solid .13 mm setlinewidth stroke\n");
	    fprintf(fp, "\n");
	}
	if (!LogX() && X0 < .0 && X1 > .0) {
	    fprintf(fp, "%%%% Y-axis\n");
	    fprintf(fp, "%9.4f %9.4f M\n", .0, Y0);
	    fprintf(fp, "%9.4f %9.4f L\n", .0, Y1);
	    fprintf(fp, "solid .13 mm setlinewidth stroke\n");
	    fprintf(fp, "\n");
	}
    }

    if (Grid()) {
	double
	    Xincr  = 50*wi->XUnitsPerPixel,
	    Yincr  = 25*wi->YUnitsPerPixel;

	fprintf(fp, "%%%% X-Unit\n"); 
	fprintf(fp, "3.8 mm /%s choose-font\n", psLabelFont); {
	    Point2d xuspot = {X1 + 13*dX, Y0}; /* X-label position */
	    if (expX != 0) {
		xuspot.x += (strlen(XUnitText) + 5)*6*dX;
		fprintf(fp, "(%s x 10) %9.4f %9.4f  %d just-string\n",
			XUnitText, xuspot.x, xuspot.y, T_RIGHT);
		fprintf(fp, "(%d)      %9.4f %9.4f  %d just-string\n",
			expX, xuspot.x, xuspot.y, T_LOWERLEFT);
	    } else {
		fprintf(fp, "(%s) %9.4f %9.4f  %d just-string\n",
			XUnitText, xuspot.x, xuspot.y, T_LEFT);
	    }
	    fprintf(fp, "%%\n");
	}

	fprintf(fp, "%%%% X-Grid\n");
	for (double x = initGrid(wi->UsrOrgX, Xincr, LogX());
	     x < wi->UsrOppX; x = stepGrid()) {
	    double
		X = x*shiftX;    /* shiftX = 1/pow(10, expX) */
	    fprintf(fp, "(%7.2f) %9.4f %9.4f  %d just-string\n",
		    X, X, Y0 - 8*dY, T_TOP);
	}
	fprintf(fp, "%%\n");
	
	fprintf(fp, "newpath\n");
	fprintf(fp, "solid .01 mm setlinewidth\n");
	for (double x = initGrid(wi->UsrOrgX, Xincr, LogX());
	     x < wi->UsrOppX; x = stepGrid()) {
	    double
		X = x*shiftX;    /* shiftX = 1/pow(10, expX) */
	    if (X == 0.0 && !LogX())
		fprintf(fp, "solid .13 mm setlinewidth\n");
	    fprintf(fp, "%9.4f %9.4f M ", X, Y0);
	    if (Ticks()) {
		const double Ticklen = 10*dY;
		fprintf(fp, "%9.4f %9.4f L stroke\n", X, Y0 + Ticklen);
		fprintf(fp, "%9.4f %9.4f M ", X, Y1 - Ticklen);
	    }
	    fprintf(fp, "%9.4f %9.4f L stroke\n", X, Y1);
	    if (X == 0.0 && !LogX())
		fprintf(fp, "solid .01 mm setlinewidth\n");
	}
	fprintf(fp, "\n");
	
	fprintf(fp, "%%%% Y-Unit\n");
	fprintf(fp, "3.8 mm /%s choose-font\n", psLabelFont); {
            Point2d yuspot = {X0 - 40*dX, Y1 + 20*dY};  /* Y-label position */
	    if (expY) {
		yuspot.x += 6*(strlen(YUnitText) + 5)*dX;
		fprintf(fp, "(%s x 10) %9.4f %9.4f  %d just-string\n",
			YUnitText, yuspot.x, yuspot.y, T_RIGHT);
		fprintf(fp, "(%d)      %9.4f %9.4f  %d just-string\n",
			expY, yuspot.x, yuspot.y, T_LOWERLEFT);
	    } else {
		yuspot.x += 6*(strlen(YUnitText))*dX;
		fprintf(fp, "(%s) %9.4f %9.4f  %d just-string\n",
			YUnitText,  yuspot.x, yuspot.y, T_RIGHT);
	    }
	    fprintf(fp, "%%\n");
	}

	fprintf(fp, "%%%% Y-Grid\n");
	for (double y = initGrid(wi->UsrOrgY, Yincr, LogY());
	     y < wi->UsrOppY; y = stepGrid()) {
	    double
		Y = y*shiftY;    /* shiftY = 1/pow(10,expY) */
	    fprintf(fp, "(%7.2f) %9.4f %9.4f  %d just-string\n",
		    Y, X0 - 4*dX, Y, T_RIGHT);
	}
	fprintf(fp, "%%\n");
	
	fprintf(fp, "newpath\n");
	fprintf(fp, "solid .01 mm setlinewidth\n");
	for (double y = initGrid(wi->UsrOrgY, Yincr, LogY());
	     y < wi->UsrOppY; y = stepGrid()) {
	    double Y = y*shiftY;    /* shiftY = 1/pow(10,expY) */
	    if (Y == 0.0 && !LogX())
		fprintf(fp, "solid .13 mm setlinewidth\n");
	    fprintf(fp, "%9.4f %9.4f M ", X0, Y);
	    if (Ticks()) {
		const double Ticklen = 10*dX;
		fprintf(fp, "%9.4f %9.4f L stroke\n", X0 + Ticklen, Y);
		fprintf(fp, "%9.4f %9.4f M ", X1 - Ticklen, Y);
	    }
	    fprintf(fp, "%9.4f %9.4f L stroke\n", X1, Y);
	    if (Y == 0.0 && !LogX()) 
		fprintf(fp, "solid .01 mm setlinewidth\n");
	}
	fprintf(fp, "\n");
    }
    
    /* legend */
    if (Grid()) {
	fprintf(fp, "%%%% Legend\n");
	fprintf(fp, "3.8 mm /%s choose-font\n", psLabelFont);
	fprintf(fp, " .29 mm setlinewidth\n");
	fprintf(fp, "newpath\n");
	for (int idx = 0; idx < MAXSETS; idx++) {
	    if (PlotData[idx].list && *PlotData[idx].setName != 0) {
		Point2d lspot= {X1 + 5*dX, Y1 - 20*idx*dY}; /* legent top */
		double lwidth = pmDouble("LegendWidth")*dX + 8*dX;
		/* Line */
		fprintf(fp, "dash%d    ", idx % 8);
		fprintf(fp, "% .4f %9.4f M ", lspot.x, lspot.y);
		if (Markers())     fprintf(fp, "mark%d ", idx % 8);
		else if (LargePixels()) fprintf(fp, "P ");
		else if (NoLines())     fprintf(fp, "p ");
		fprintf(fp, "%9.4f %9.4f L ",  lspot.x + lwidth, lspot.y);
		if (!NoLines()) fprintf(fp, "stroke\n");
		/* SetName */
		fprintf(fp, "(%-7s) %.4f %9.4f  %d just-string\n",
			PlotData[idx].setName,
			lspot.x, lspot.y - 2.5*dY, T_UPPERLEFT);
		fprintf(fp, "%%\n");
	    }
	}
	fprintf(fp, "\n");
    }

    fprintf(fp, "%%%% BoundBox and Clipping\n");
    fprintf(fp, "newpath\n");
    fprintf(fp, "%9.4f %9.4f M\n", X0, Y0);
    fprintf(fp, "%9.4f %9.4f L\n", X1, Y0);
    fprintf(fp, "%9.4f %9.4f L\n", X1, Y1);
    fprintf(fp, "%9.4f %9.4f L\n", X0, Y1);
    fprintf(fp, "closepath\n");
    if (BoundBox()) {
	fprintf(fp, "gsave\n");
	fprintf(fp, "solid .01 mm setlinewidth stroke\n");
	fprintf(fp, "grestore\n");
    }
    fprintf(fp, "clip\n");
    fprintf(fp, "\n");

    for (int idx = 0; idx < MAXSETS; idx++) {
	PointList *list = PlotData[idx].list;
	if (PlotData[idx].list)
	    fprintf(fp, "%%%% Set %d: %s\n", idx, PlotData[idx].setName);
	for ( ; list; list = list->next) {
	    int	N = list->numPoints;
	    if (N > 2) { fprintf(fp, "%%%% %d points\n", N); }
	    if (list == PlotData[idx].list && !NoLines()) {
		fprintf(fp, "dash%d .29 mm setlinewidth\n",idx % 8);
		fprintf(fp, "newpath\n");
	    }
	    for (int i = 0; i < N; i++) {
		fprintf(fp, "%9.4f %9.4f %s",
			list->pt[i].x*shiftX, list->pt[i].y*shiftY,
			(i == 0) ? "M" : "L");
		if (Markers())
		    fprintf(fp, " mark%d", idx % 8);
		else if (LargePixels())
		    fprintf(fp, " P");
		else if (NoLines())
		    fprintf(fp, " p");
		fprintf(fp, "%s", (N > 2) ? "\n" : " ");
	    }
	    if (!NoLines()) fprintf(fp, "stroke\n");
	}
	if (PlotData[idx].list) {
	    fprintf(fp, "%%%% /Set %d\n\n", idx);
	}
    }
    fprintf(fp, "showpage\n");
    fprintf(fp, "%%%% End\n");
    fclose(fp);
}
